from fw_ddsm.community import *
from fw_ddsm.aggregator import *

class Experiment:

    def __init__(self, algorithms, num_households):
        self.algorithms = algorithms
        self.num_households = num_households
        self.num_iteration = 0
        self.num_intervals = 0

    def new_data(self, file_demand_list=file_demand_list,
                 file_pricing_table=file_pricing_table,
                 data_folder="test2",
                 file_probability=file_probability,
                 max_demand_multiplier=maxium_demand_multiplier,
                 num_tasks_dependent=no_tasks_dependent,
                 full_flex_task_min=no_full_flex_tasks_min, full_flex_task_max=0,
                 semi_flex_task_min=no_semi_flex_tasks_min, semi_flex_task_max=0,
                 fixed_task_min=no_fixed_tasks_min, fixed_task_max=0,
                 inconvenience_cost_weight=care_f_weight, max_care_factor=care_f_max):
        self.data_folder = data_folder

        # 1. generate new households, trackers and a pricing table
        self.community = Community()
        self.community.new(num_households=self.num_households,
                           file_probability_path=file_probability,
                           file_demand_list_path=file_demand_list,
                           algorithms_options=self.algorithms,
                           write_to_file_path=self.data_folder,
                           max_demand_multiplier=max_demand_multiplier,
                           num_tasks_dependent=num_tasks_dependent,
                           full_flex_task_min=full_flex_task_min, full_flex_task_max=full_flex_task_max,
                           semi_flex_task_min=semi_flex_task_min, semi_flex_task_max=semi_flex_task_max,
                           fixed_task_min=fixed_task_min, fixed_task_max=fixed_task_max,
                           inconvenience_cost_weight=inconvenience_cost_weight, max_care_factor=max_care_factor)

        self.aggregator = Aggregator()
        self.aggregator.new(normalised_pricing_table_csv=file_pricing_table,
                            aggregate_preferred_demand_profile=self.community.aggregate_data[k0_demand],
                            algorithms_options=self.algorithms, write_to_file_path=self.data_folder)

        print("Households and the aggregator are created. ")

    def read_data(self, read_from_folder="test2"):
        self.community = Community()
        self.community.read(read_from_file=read_from_folder)
        self.aggregator = Aggregator()
        self.aggregator.read(read_from_file=read_from_folder)

        print("Households and the aggregator are read. ")

    def iteration(self, algorithm):
        scheduling_method = algorithm[k2_before_fw]
        pricing_method = algorithm[k2_after_fw]

        # aggregator, k = 0
        step = 1
        aggregate_demand_profile = self.aggregator.aggregator[pricing_method][k0_demand][0]
        prices, consumption_cost = self.aggregator.prices_and_cost(aggregate_demand_profile=aggregate_demand_profile)
        self.aggregator.update(num_iteration=0, pricing_method=pricing_method,
                               step=step, demands=aggregate_demand_profile, prices=prices,
                               consumption_cost=consumption_cost,
                               inconvenience_cost=0)

        num_iteration = 1
        while step > 0:
            # community, k > 0
            aggregate_demand_profile_new, total_inconvenience, time_scheduling_iteration \
                = self.community.schedule_all(num_iteration=num_iteration, prices=prices,
                                              scheduling_method=scheduling_method)
            self.community.update_aggregate_data(num_iteration=num_iteration, demands=aggregate_demand_profile_new,
                                                 penalty=total_inconvenience, time=time_scheduling_iteration,
                                                 scheduling_method=scheduling_method)

            # aggregator, k > 0
            aggregate_demand_profile_fw, step, prices, total_consumption_cost_fw, total_inconvenience_fw \
                = self.aggregator.find_step_size(num_iteration=num_iteration,
                                                 demand_profile=aggregate_demand_profile_new,
                                                 inconvenience=total_inconvenience,
                                                 pricing_method=pricing_method)
            self.aggregator.update(num_iteration=num_iteration, pricing_method=pricing_method,
                                   step=step, prices=prices, consumption_cost=total_consumption_cost_fw,
                                   demands=aggregate_demand_profile_fw, inconvenience_cost=total_inconvenience_fw)

            num_iteration += 1

        print(f"Converged in {num_iteration - 1}")
        #
        #
        #
        #     num_iteration += 1

    def final_schedule(self, algorithm):
        scheduling_method = algorithm[k2_before_fw]
        pricing_method = algorithm[k2_after_fw]

        start_time_probability_distribution = self.aggregator.compute_start_time_probabilities(pricing_method)
        final_aggregate_demand_profile, final_total_inconvenience \
            = self.community.decide_final_schedules(scheduling_method=scheduling_method,
                                                    probability_distribution=start_time_probability_distribution)

        final_prices, final_consumption_cost = self.aggregator.prices_and_cost(final_aggregate_demand_profile)
        self.aggregator.update(num_iteration=None, final=True, pricing_method=pricing_method,
                               demands=final_aggregate_demand_profile, prices=final_prices,
                               consumption_cost=final_consumption_cost, inconvenience_cost=final_total_inconvenience)

        print(self.aggregator.aggregator[pricing_method][k0_final][k0_demand])
        print(final_prices)
        print(f"Preferred cost is {self.aggregator.aggregator[pricing_method][k0_cost][0]}, "
              f"PAR is {self.aggregator.aggregator[pricing_method][k0_par][0]}")
        print(f"Final cost is {final_consumption_cost}, "
              f"PAR is {self.aggregator.aggregator[pricing_method][k0_final][k0_par]} and "
              f"inconvenience is {final_total_inconvenience}.")

        return final_aggregate_demand_profile, final_total_inconvenience



from multiprocessing import Pool, cpu_count
import pickle
from fw_ddsm.cfunctions import average
from fw_ddsm.household import *


class Community:

    def __init__(self, num_intervals=no_intervals, num_periods=no_periods):
        self.num_intervals = num_intervals
        self.num_periods = num_periods
        self.num_intervals_periods = int(num_intervals / num_periods)

    def read(self, read_from_file, inconvenience_cost_weight=None):
        read_from_file = read_from_file if read_from_file.endswith("/") \
            else read_from_file + "/"
        self.households, self.aggregate_data = self.__existing_households(file_path=read_from_file,
                                                                          inconvenience_cost_weight=inconvenience_cost_weight)
        self.num_households = len(self.households)

    def new(self, file_probability_path, file_demand_list_path, algorithms_options,
            num_households=no_households,
            max_demand_multiplier=maxium_demand_multiplier,
            num_tasks_dependent=no_tasks_dependent,
            full_flex_task_min=no_full_flex_tasks_min, full_flex_task_max=0,
            semi_flex_task_min=no_semi_flex_tasks_min, semi_flex_task_max=0,
            fixed_task_min=no_fixed_tasks_min, fixed_task_max=0,
            inconvenience_cost_weight=care_f_weight, max_care_factor=care_f_max,
            write_to_file_path=None):

        self.num_households = num_households

        self.households, self.aggregate_data \
            = self.__new_households(file_probability_path,
                                    file_demand_list_path,
                                    algorithms_options,
                                    max_demand_multiplier=max_demand_multiplier,
                                    num_tasks_dependent=num_tasks_dependent,
                                    full_flex_task_min=full_flex_task_min,
                                    full_flex_task_max=full_flex_task_max,
                                    semi_flex_task_min=semi_flex_task_min,
                                    semi_flex_task_max=semi_flex_task_max,
                                    fixed_task_min=fixed_task_min,
                                    fixed_task_max=fixed_task_max,
                                    inconvenience_cost_weight=inconvenience_cost_weight,
                                    max_care_factor=max_care_factor,
                                    write_to_file_path=write_to_file_path)

    def update_household(self, key, num_iteration, scheduling_method, demands=None, starts=None, penalty=None):

        if demands is not None:
            self.households[key][scheduling_method][k0_demand][num_iteration] = demands
            self.households[key][scheduling_method][k0_demand_max][num_iteration] = max(demands)
            # self.households[key][scheduling_method][k0_demand_total][num_iteration] = sum(demands)
            # self.households[key][scheduling_method][k0_par][num_iteration] = max(demands) / average(demands)
        if penalty is not None:
            self.households[key][scheduling_method][k0_penalty][num_iteration] = penalty
        # if time is not None:
        #     self.households[key][scheduling_method][k0_time][num_iteration] = time
        if starts is not None:
            self.households[key][scheduling_method][k0_starts][num_iteration] = starts

    def update_aggregate_data(self, num_iteration, scheduling_method, demands=None, prices=None, penalty=None,
                              time=None):

        if demands is not None:
            self.aggregate_data[scheduling_method][k0_demand][num_iteration] = demands
            self.aggregate_data[scheduling_method][k0_demand_max][num_iteration] = max(demands)
            self.aggregate_data[scheduling_method][k0_demand_total][num_iteration] = sum(demands)
            self.aggregate_data[scheduling_method][k0_par][num_iteration] = max(demands) / average(demands)
        if penalty is not None:
            self.aggregate_data[scheduling_method][k0_penalty][num_iteration] = penalty
        if time is not None:
            self.aggregate_data[scheduling_method][k0_time][num_iteration] = time
        if prices is not None:
            self.aggregate_data[scheduling_method][k0_prices][num_iteration] = prices

    def schedule_all(self, num_iteration, prices, scheduling_method, model=None, solver=None, search=None):
        num_periods = len(prices)
        num_intervals_period = int(self.num_intervals / num_periods)
        if num_periods != self.num_intervals:
            prices = [p for p in prices for _ in range(num_intervals_period)]
        else:
            prices = [p for p in prices]
        self.update_aggregate_data(num_iteration=num_iteration, scheduling_method=scheduling_method, prices=prices)

        households = self.households
        print(f"{num_iteration}. Start scheduling households using {scheduling_method}...")
        pool = Pool(cpu_count())
        results = pool.starmap_async(self.schedule_household,
                                     [(household, prices, scheduling_method, model, solver, search)
                                      for household in households.values()]).get()
        pool.close()
        pool.join()

        num_intervals = self.num_intervals
        aggregate_demand_profile = [0] * num_intervals
        total_inconvenience = 0
        time_scheduling_iteration = 0
        for res in results:
            key = res[h_key]
            demands_household = res[k0_demand]
            penalty_household = res[k0_penalty]
            time_household = res[k0_time]
            self.update_household(key=key, num_iteration=num_iteration, starts=res[k0_starts],
                                  penalty=penalty_household, demands=demands_household,
                                  scheduling_method=scheduling_method)
            aggregate_demand_profile = [x + y for x, y in zip(demands_household, aggregate_demand_profile)]
            total_inconvenience += penalty_household
            time_scheduling_iteration += time_household

        return aggregate_demand_profile, total_inconvenience, time_scheduling_iteration

    def schedule_household(self, household, prices, scheduling_method, model, solver, search):
        existing_household = Household()
        result = existing_household.schedule(prices=prices, scheduling_method=scheduling_method,
                                             household=household, model=model, solver=solver, search=search)
        return result

    def decide_final_schedules(self, scheduling_method, probability_distribution):
        existing_household = Household()
        final_aggregate_demand_profile = [0] * self.num_intervals
        final_total_inconvenience = 0
        for household in self.households.values():
            chosen_demand_profile, chosen_penalty \
                = existing_household.decide_final_schedule(household=household, scheduling_method=scheduling_method,
                                                           probability_distribution=probability_distribution)
            final_aggregate_demand_profile = [x + y for x, y in
                                              zip(chosen_demand_profile, final_aggregate_demand_profile)]
            final_total_inconvenience += chosen_penalty

        self.aggregate_data[scheduling_method][k0_final][k0_demand] = final_aggregate_demand_profile
        self.aggregate_data[scheduling_method][k0_final][k0_penalty] = final_total_inconvenience

        return final_aggregate_demand_profile, final_total_inconvenience

    def __new_households(self, file_probability_path, file_demand_list_path, algorithms_options,
                         max_demand_multiplier=maxium_demand_multiplier,
                         num_tasks_dependent=no_tasks_dependent,
                         full_flex_task_min=no_full_flex_tasks_min, full_flex_task_max=0,
                         semi_flex_task_min=no_semi_flex_tasks_min, semi_flex_task_max=0,
                         fixed_task_min=no_fixed_tasks_min, fixed_task_max=0,
                         inconvenience_cost_weight=care_f_weight, max_care_factor=care_f_max,
                         write_to_file_path=None):
        # ---------------------------------------------------------------------- #

        # ---------------------------------------------------------------------- #

        num_households = self.num_households
        preferred_demand_profile = genfromtxt(file_probability_path, delimiter=',', dtype="float")
        list_of_devices_power = genfromtxt(file_demand_list_path, delimiter=',', dtype="float")

        households = dict()
        num_intervals = self.num_intervals
        aggregate_demand_profile = [0] * num_intervals

        for h in range(num_households):
            new_household = Household()
            new_household.new(preferred_demand_profile,
                              list_of_devices_power,
                              algorithms_options,
                              max_demand_multiplier=max_demand_multiplier,
                              num_tasks_dependent=num_tasks_dependent,
                              full_flex_task_min=full_flex_task_min,
                              full_flex_task_max=full_flex_task_max,
                              semi_flex_task_min=semi_flex_task_min,
                              semi_flex_task_max=semi_flex_task_max,
                              fixed_task_min=fixed_task_min,
                              fixed_task_max=fixed_task_max,
                              inconvenience_cost_weight=inconvenience_cost_weight,
                              max_care_factor=max_care_factor, id=h)
            household_profile = new_household.tasks[h_demand_profile]
            aggregate_demand_profile = [x + y for x, y in zip(household_profile, aggregate_demand_profile)]
            households[h] = new_household.tasks.copy()

        # create aggregate trackers
        max_demand = max(aggregate_demand_profile)
        total_demand = sum(aggregate_demand_profile)
        par = max_demand / average(aggregate_demand_profile)
        aggregate_data = dict()
        aggregate_data[k0_demand] = aggregate_demand_profile
        for algorithm in algorithms_options.values():
            for alg in algorithm.values():
                if "fw" not in alg:
                    aggregate_data[alg] = dict()
                    aggregate_data[alg][k0_demand] = dict()
                    aggregate_data[alg][k0_demand_max] = dict()
                    aggregate_data[alg][k0_demand_total] = dict()
                    aggregate_data[alg][k0_par] = dict()
                    aggregate_data[alg][k0_penalty] = dict()
                    aggregate_data[alg][k0_final] = dict()
                    aggregate_data[alg][k0_prices] = dict()
                    aggregate_data[alg][k0_cost] = dict()
                    aggregate_data[alg][k0_time] = dict()

                    aggregate_data[alg][k0_demand][0] = aggregate_demand_profile
                    aggregate_data[alg][k0_demand_max][0] = max_demand
                    aggregate_data[alg][k0_demand_total][0] = total_demand
                    aggregate_data[alg][k0_par][0] = par
                    aggregate_data[alg][k0_par][0] = par
                    aggregate_data[alg][k0_penalty][0] = 0
                    aggregate_data[alg][k0_cost][0] = 0
                    aggregate_data[alg][k0_time][0] = 0

        # write household data and area data into files
        if write_to_file_path is not None:
            write_to_file_path = write_to_file_path if write_to_file_path.endswith("/") \
                else write_to_file_path + "/"
            path = Path(write_to_file_path)
            if not path.exists():
                path.mkdir(mode=0o777, parents=True, exist_ok=False)

            with open(f"{write_to_file_path}{file_community_pkl}", 'wb+') as f:
                pickle.dump(households, f, pickle.HIGHEST_PROTOCOL)
            f.close()

            with open(f"{write_to_file_path}{file_community_meta_pkl}", 'wb+') as f:
                pickle.dump(aggregate_data, f, pickle.HIGHEST_PROTOCOL)
            f.close()

        return households, aggregate_data

    def __existing_households(self, file_path, inconvenience_cost_weight=None):
        # ---------------------------------------------------------------------- #
        # ---------------------------------------------------------------------- #

        with open(f"{file_path}{file_community_pkl}", 'rb') as f:
            households = pickle.load(f)
        f.close()

        with open(f"{file_path}{file_community_meta_pkl}", 'rb') as f:
            households_meta = pickle.load(f)
        f.close()

        if inconvenience_cost_weight is not None:
            for household in households.values():
                household["care_factor_weight"] = inconvenience_cost_weight

        return households, households_meta



import random as r
import numpy as np
from numpy import sqrt, pi, random
from numpy.random import choice
from json import dumps, load
from pathlib import Path
import timeit
from datetime import timedelta
from minizinc import *
from numpy import genfromtxt
from fw_ddsm.parameter import *


class Household:

    def __init__(self, num_intervals=no_intervals, num_periods=no_periods):
        self.num_intervals = num_intervals
        self.num_periods = num_periods
        self.num_intervals_periods = int(num_intervals / num_periods)
        self.tasks = dict()

    def read(self, read_from_file=None, existing_household=None):
        self.tasks = self.__existing_household(household_file=read_from_file,
                                               existing_household=existing_household)
        if existing_household is None:
            print(f"Household{self.tasks[h_key]} is read.")

    def new(self, preferred_demand_profile, list_of_devices_power, algorithms_options,
            preferred_demand_profile_csv=None,
            max_demand_multiplier=maxium_demand_multiplier,
            num_tasks_dependent=no_tasks_dependent,
            full_flex_task_min=no_full_flex_tasks_min, full_flex_task_max=0,
            semi_flex_task_min=no_semi_flex_tasks_min, semi_flex_task_max=0,
            fixed_task_min=no_fixed_tasks_min, fixed_task_max=0,
            inconvenience_cost_weight=care_f_weight, max_care_factor=care_f_max,
            write_to_file_path=None, id=0):

        if preferred_demand_profile_csv is not None:
            preferred_demand_profile = genfromtxt(preferred_demand_profile_csv, delimiter=',', dtype="float")

        self.tasks = self.__new_household(preferred_demand_profile,
                                          list_of_devices_power,
                                          algorithms_options,
                                          max_demand_multiplier=max_demand_multiplier,
                                          num_tasks_dependent=num_tasks_dependent,
                                          full_flex_task_min=full_flex_task_min,
                                          full_flex_task_max=full_flex_task_max,
                                          semi_flex_task_min=semi_flex_task_min,
                                          semi_flex_task_max=semi_flex_task_max,
                                          fixed_task_min=fixed_task_min,
                                          fixed_task_max=fixed_task_max,
                                          inconvenience_cost_weight=inconvenience_cost_weight,
                                          max_care_factor=max_care_factor,
                                          write_to_file_path=write_to_file_path, id=id)
        # print(f"Household{self.data[h_key]} is created.")

    def schedule(self, prices, scheduling_method, household=None, model=None, solver=None, search=None):

        def preprocessing():
            max_duration = max(durations)
            # this big cost and big cost * number_tasks need to be smaller than the largest number that the solver can handle
            big_value = max_demand * max_duration * prices[-2] + \
                        inconvenience_cost_weight * max_care_factor * num_intervals
            objective_value_matrix = []
            for power, pst, est, lft, dur, cf in zip(powers, preferred_starts, earliest_starts,
                                                     latest_ends, durations, care_factors):
                objective_value_task = []
                for t in range(num_intervals):
                    if est <= t <= lft - dur + 1:
                        rc = abs(t - pst) * cf * inconvenience_cost_weight
                        try:
                            rc += sum([prices[j % num_intervals] for j in range(t, t + dur)]) * power
                        except IndexError:
                            print("Error: check the prices.")
                    else:
                        rc = big_value
                    objective_value_task.append(int(rc))
                objective_value_matrix.append(objective_value_task)
            return objective_value_matrix, big_value

        # read tasks
        if household is None:
            household = self.tasks
        key = household[h_key]
        powers = household[h_powers]
        durations = household[h_durs]
        earliest_starts = household[h_ests]
        latest_ends = household[h_lfs]
        preferred_starts = household[h_psts]
        care_factors = household[h_cfs]
        max_care_factor = household[h_max_cf]
        precedents = [x[0] for x in list(household[h_precs].values())]
        successors = list(household[h_precs].keys())
        succ_delays = household[h_succ_delay]   # need to change this format when sending it to the solver
        no_precedents = household[h_no_precs]
        max_demand = household[h_max_demand]
        inconvenience_cost_weight = household[h_incon_weight]

        # read prices
        num_intervals = len(household[h_demand_profile])
        num_periods = len(prices)
        num_intervals_period = int(num_intervals / num_periods)
        if num_periods != num_intervals:
            prices = [int(p * 10) for p in prices for _ in range(num_intervals_period)]
        else:
            prices = [int(p * 10) for p in prices]

        # begin scheduling
        objective_values, big_value = preprocessing()
        if "minizinc" in scheduling_method:
            model = file_cp_pre if model is None else model
            solver = solver_name if solver is None else solver
            search = f"int_search(actual_starts, {variable_selection}, {value_choice}, complete)" \
                if search is None else search
            succ_delays = [x[0] for x in list(household[h_succ_delay].values())]
            actual_starts, household_demand_profile, time_scheduling \
                = self.__minizinc_model(model_file=model, solver=solver, search=search,
                                        objective_values=objective_values, powers=powers, max_demand=max_demand,
                                        durations=durations, earliest_starts=earliest_starts,
                                        preferred_starts=preferred_starts,
                                        latest_ends=latest_ends, successors=successors, precedents=precedents,
                                        no_precedents=no_precedents, succ_delays=succ_delays, care_factors=care_factors,
                                        prices=prices, inconvenience_cost_weight=inconvenience_cost_weight)
        else:
            actual_starts, household_demand_profile, time_scheduling \
                = self.__ogsa(objective_values=objective_values, big_value=big_value,
                              powers=powers, durations=durations, preferred_starts=preferred_starts,
                              latest_ends=latest_ends, max_demand=max_demand,
                              successors=successors, precedents=precedents, succ_delays=succ_delays,
                              randomness=False)

        # return results
        penalty_household = sum([abs(pst - ast) * cf for pst, ast, cf
                                 in zip(preferred_starts, actual_starts, care_factors)]) * inconvenience_cost_weight

        return {h_key: key, k0_demand: household_demand_profile, k0_starts: actual_starts,
                k0_penalty: penalty_household, k0_time: time_scheduling}


    def update(self, num_iteration, scheduling_method, starts=None, demands=None, penalty=None, time=None):
        if starts is not None:
            self.tasks[scheduling_method][k0_starts][num_iteration] = starts
        if demands is not None:
            self.tasks[scheduling_method][k0_demand][num_iteration] = demands
        if penalty is not None:
            self.tasks[scheduling_method][k0_penalty][num_iteration] = penalty
        if time is not None:
            self.tasks[scheduling_method][k0_time][num_iteration] = time


    def decide_final_schedule(self, scheduling_method, probability_distribution, household=None):

        if household is None:
            household = self.tasks
        chosen_iter = choice(len(probability_distribution), size=1, p=probability_distribution)[0]
        chosen_demand_profile = household[scheduling_method][k0_demand][chosen_iter]
        chosen_penalty = household[scheduling_method][k0_penalty][chosen_iter] * household[h_incon_weight]

        if household is None:
            self.tasks[scheduling_method][k0_final][k0_demand] = chosen_demand_profile
            self.tasks[scheduling_method][k0_final][k0_penalty] = chosen_penalty


        return chosen_demand_profile, chosen_penalty

    def __new_task(self, mode_value, list_of_devices_power, pst_probabilities, max_care_factor,
                   scheduling_window_width, ):
        # ---------------------------------------------------------------------- #
        # mode_value:
        #       a parameter for generation the duration using the Rayleigh distribution
        # list_of_device_power:
        #       the list of commonly used devices' power, or the list of devices' power rates
        # pst_probabilities:
        #       the probability distribution used for sampling the PST
        # max_care_factor:
        #       the maximum care factor, e.g. 1, or 10 or 100
        # scheduling_window_width:
        #       "full", can be rescheduled to any time intervals;
        #       "fixed", cannot be moved;
        # ---------------------------------------------------------------------- #

        num_intervals = self.num_intervals
        num_periods = self.num_periods
        num_intervals_periods = self.num_intervals_periods

        # task power
        power = r.choice(list_of_devices_power)
        power = int(power * 1000)

        # task duration
        duration = max(1, int(random.rayleigh(mode_value, 1)[0]))

        # task preferred start time
        preferred_start_time = max(int(np.random.choice(a=num_periods, size=1, p=pst_probabilities)[0]) *
                                   num_intervals_periods + r.randint(-num_intervals_periods + 1, num_intervals_periods),
                                   0)
        preferred_start_time = min(preferred_start_time, num_intervals - 1)

        # task earliest starting time and latest finish time
        if scheduling_window_width == "full":
            earliest_start_time = 0
            latest_finish_time = num_intervals - 1 + duration
        elif scheduling_window_width == "fixed":
            earliest_start_time = preferred_start_time
            latest_finish_time = preferred_start_time + duration - 1
        else:
            earliest_start_time = r.randint(0, preferred_start_time)
            latest_finish_time = num_intervals - 1 + duration

        # task care factor
        care_factor = int(r.choice([i for i in range(1, max_care_factor + 1)]))

        return power, duration, preferred_start_time, earliest_start_time, latest_finish_time, care_factor

    def __new_household(self, preferred_demand_profile, list_of_devices_power, algorithms_options,
                        max_demand_multiplier=maxium_demand_multiplier,
                        num_tasks_dependent=no_tasks_dependent,
                        full_flex_task_min=no_full_flex_tasks_min, full_flex_task_max=0,
                        semi_flex_task_min=no_semi_flex_tasks_min, semi_flex_task_max=0,
                        fixed_task_min=no_fixed_tasks_min, fixed_task_max=0,
                        inconvenience_cost_weight=care_f_weight, max_care_factor=care_f_max,
                        write_to_file_path=None, id=None):
        # ---------------------------------------------------------------------- #
        # preferred_demand_profile:
        #       the demand profile used for computing the probability distribution for sampling the preferred start times
        # list_of_devices_power:
        #       the power of the tasks
        # max_demand_multiplier:
        # num_tasks_dependent:
        # full_flex_task_min and full_flex_task_max:
        #       the minimum and the maximum number of fully flexible tasks
        # semi_flex_task_min and semi_flex_task_max:
        #       the minimum and the maximum number of semi-flexible tasks
        # fixed_task_min and care_f_weight:
        # inconvenience_cost_weight and max_care_factor:
        # num_intervals, num_periods and num_intervals_periods:
        # write_to_file:
        #       whether to write the created household data to a file
        # ---------------------------------------------------------------------- #

        num_intervals = self.num_intervals
        pst_probabilities = [int(p) for p in preferred_demand_profile]
        sum_pst_probabilities = sum(pst_probabilities)
        pst_probabilities = [p / sum_pst_probabilities for p in pst_probabilities]

        # I meant mean value is 40 minutes
        mean_value = 40.0 / (24.0 * 60.0 / num_intervals)
        mode_value = sqrt(2 / pi) * mean_value

        # task details
        preferred_starts = []
        earliest_starts = []
        latest_ends = []
        durations = []
        powers = []
        care_factors = []
        household_demand_profile = [0] * num_intervals

        # tasks in the household
        def get_new_tasks(num_tasks, scheduling_window_width):
            for counter_j in range(num_tasks):
                demand, duration, p_start, e_start, l_finish, care_f \
                    = self.__new_task(mode_value, list_of_devices_power,
                                      pst_probabilities, max_care_factor, scheduling_window_width)

                powers.append(demand)
                durations.append(duration)
                preferred_starts.append(p_start)
                earliest_starts.append(e_start)
                latest_ends.append(l_finish)
                care_factors.append(care_f)
                # add this task demand to the household demand
                for d in range(duration):
                    household_demand_profile[(p_start + d) % num_intervals] += demand

        fixed_task_max = max(fixed_task_max, fixed_task_min)
        semi_flex_task_max = max(semi_flex_task_max, semi_flex_task_min)
        full_flex_task_max = max(full_flex_task_max, full_flex_task_min)

        num_fixed_tasks = r.randint(fixed_task_min, fixed_task_max)
        num_semi_flex_tasks = r.randint(semi_flex_task_min, semi_flex_task_max)
        num_full_flex_tasks = r.randint(full_flex_task_min, full_flex_task_max)

        get_new_tasks(num_fixed_tasks, "fixed")
        get_new_tasks(num_semi_flex_tasks, "semi")
        get_new_tasks(num_full_flex_tasks, "full")

        # set the household demand limit
        maximum_demand = max(powers) * max_demand_multiplier

        # precedence among tasks
        precedors = dict()
        no_precedences = 0
        succ_delays = dict()

        def retrieve_precedes(list0):
            list3 = []
            for l in list0:
                if l in precedors:
                    list2 = precedors[l]
                    retrieved_list = retrieve_precedes(list2)
                    list3.extend(retrieved_list)
                else:
                    list3.append(l)
            return list3

        def add_precedes(task, previous, delay):
            if task not in precedors:
                precedors[task] = [previous]
                succ_delays[task] = [delay]
            else:
                precedors[task].append(previous)
                succ_delays[task].append(delay)

        num_total_tasks = num_full_flex_tasks + num_semi_flex_tasks + num_fixed_tasks
        for t in range(num_total_tasks - num_tasks_dependent, num_total_tasks):
            if r.choice([True, False]):
                previous_tasks = list(range(t))
                r.shuffle(previous_tasks)
                for prev in previous_tasks:
                    if preferred_starts[prev] + durations[prev] - 1 < preferred_starts[t] \
                            and earliest_starts[prev] + durations[prev] < latest_ends[t] - durations[t] + 1:

                        if prev not in precedors:
                            # feasible delay
                            succeding_delay = num_intervals - 1
                            add_precedes(t, prev, succeding_delay)
                            no_precedences += 1
                            break
                        else:
                            # find all precedors of this previous task
                            precs_prev = retrieve_precedes([prev])
                            precs_prev.append(prev)

                            precs_prev_duration = sum([durations[x] for x in precs_prev])
                            latest_pstart = preferred_starts[precs_prev[0]]
                            latest_estart = earliest_starts[precs_prev[0]]

                            if latest_pstart + precs_prev_duration - 1 < preferred_starts[t] \
                                    and latest_estart + precs_prev_duration < latest_ends[t] - durations[t] + 1:
                                succeding_delay = num_intervals - 1
                                add_precedes(t, prev, succeding_delay)
                                no_precedences += 1
                                break

        household = dict()
        if id is not None:
            household[h_key] = id
        household[h_psts] = preferred_starts
        household[h_ests] = earliest_starts
        household[h_lfs] = latest_ends
        household[h_durs] = durations
        household[h_powers] = powers
        household[h_cfs] = care_factors
        household[h_max_cf] = max_care_factor
        household[h_no_precs] = no_precedences
        household[h_precs] = precedors
        household[h_succ_delay] = succ_delays
        household[h_max_demand] = maximum_demand
        household[h_demand_profile] = household_demand_profile
        household[h_incon_weight] = inconvenience_cost_weight

        for k in algorithms_options.keys():
            household[k] = dict()
            household[k][k0_starts] = dict()
            household[k][k0_penalty] = dict()
            household[k][k0_final] = dict()
            household[k][k0_demand] = dict()
            household[k][k0_demand_max] = dict()

            household[k][k0_starts][0] = household[h_psts]
            household[k][k0_penalty][0] = 0
            household[k][k0_demand][0] = household_demand_profile
            household[k][k0_demand_max][0] = max(household_demand_profile)

        if write_to_file_path is not None:
            write_to_file_path = write_to_file_path if write_to_file_path.endswith("/") \
                else write_to_file_path + "/"
            path = Path(write_to_file_path)
            if not path.exists():
                path.mkdir(mode=0o777, parents=True, exist_ok=False)
            with open(f"{write_to_file_path}household{id}.txt", "w") as f:
                f.write(dumps(household, indent=1))
            f.close()
            print(f"{write_to_file_path}household{id}.txt written.")

        return household

    def __existing_household(self, household_file=None, existing_household=None):
        # ---------------------------------------------------------------------- #
        # ---------------------------------------------------------------------- #
        if household_file is not None:
            with open(household_file, 'r') as f:
                household = load(f)
            f.close()
        else:
            household = existing_household.copy()

        return household

    def __minizinc_model(self, model_file, solver, search,
                         objective_values, powers, max_demand, durations,
                         earliest_starts, preferred_starts, latest_ends,
                         successors, precedents, no_precedents, succ_delays,
                         care_factors, prices, inconvenience_cost_weight):
        num_intervals = self.num_intervals

        # problem model
        model = Model(model_file)
        gecode = Solver.lookup(solver)
        model.add_string("solve ")
        if "gecode" in solver:
            model.add_string(":: {} ".format(search))
        model.add_string("minimize obj;")

        ins = Instance(gecode, model)
        num_tasks = len(powers)
        ins["num_intervals"] = num_intervals
        ins["num_tasks"] = num_tasks
        ins["durations"] = durations
        ins["demands"] = powers
        ins["num_precedences"] = no_precedents
        ins["predecessors"] = [p + 1 for p in precedents]
        ins["successors"] = [s + 1 for s in successors]
        ins["prec_delays"] = succ_delays
        ins["max_demand"] = max_demand

        if "ini" in model_type.lower():
            ins["prices"] = prices
            ins["preferred_starts"] = [ps + 1 for ps in preferred_starts]
            ins["earliest_starts"] = [es + 1 for es in earliest_starts]
            ins["latest_ends"] = [le + 1 for le in latest_ends]
            ins["care_factors"] = [cf * inconvenience_cost_weight for cf in care_factors]
        else:
            ins["run_costs"] = objective_values

        # solve problem model
        result = ins.solve(timeout=timedelta(seconds=2))
        # result = ins.solve()

        # process problem solution
        # obj = result.objective
        solution = result.solution.actual_starts
        if "cp" in solver_type:
            actual_starts = [int(a) - 1 for a in solution]
        else:  # "mip" in solver_type:
            actual_starts = [sum([i * int(v) for i, v in enumerate(row)]) for row in solution]
        time = result.statistics["time"].total_seconds()

        optimal_demand_profile = [0] * num_intervals
        for power, duration, a_start, i in zip(powers, durations, actual_starts, range(num_tasks)):
            for t in range(a_start, a_start + duration):
                optimal_demand_profile[t % num_intervals] += power

        return actual_starts, optimal_demand_profile, time

    def __ogsa(self, objective_values, big_value, powers, durations, preferred_starts, latest_ends, max_demand,
               successors, precedents, succ_delays, randomness=True):
        start_time = timeit.default_timer()
        num_intervals = self.num_intervals

        def retrieve_successors_or_precedents(list0, prec_or_succ_list1, succ_prec_list2):
            list_r = []
            for l in list0:
                if l in prec_or_succ_list1:
                    succ_or_prec_indices = [i2 for i2, k in enumerate(prec_or_succ_list1) if k == l]
                    succ_or_prec = [succ_prec_list2[i2] for i2 in succ_or_prec_indices]
                    succ_succ_or_prec_prec \
                        = retrieve_successors_or_precedents(succ_or_prec, prec_or_succ_list1, succ_prec_list2)
                    list_r.extend(succ_succ_or_prec_prec)
                else:
                    list_r.append(l)
            return list_r

        def check_if_successors_or_precedents_exist(checked_task_id, prec_or_succ1, succ_or_prec2):
            succs_succs_or_precs_precs = []
            if checked_task_id in prec_or_succ1:
                indices = [i2 for i2, k in enumerate(prec_or_succ1) if k == checked_task_id]
                succs_or_precs = [succ_or_prec2[i2] for i2 in indices]
                succs_succs_or_precs_precs = retrieve_successors_or_precedents(succs_or_precs, prec_or_succ1,
                                                                               succ_or_prec2)

            return succs_succs_or_precs_precs

        actual_starts = []
        household_profile = [0] * num_intervals
        num_tasks = len(powers)
        for task_id in range(num_tasks):
            power = powers[task_id]
            duration = durations[task_id]
            task_costs = objective_values[task_id]

            # if i has successors
            tasks_successors = check_if_successors_or_precedents_exist(task_id, precedents, successors)
            earliest_suc_lstart_w_delay = 0
            earliest_suc_lstart = num_intervals - 1
            if bool(tasks_successors):
                suc_durations = [durations[i2] for i2 in tasks_successors]
                suc_lends = [latest_ends[i2] for i2 in tasks_successors]
                earliest_suc_lstart = min([lend - dur for lend, dur in zip(suc_lends, suc_durations)])
                # earliest_suc_lstart_w_delay = earliest_suc_lstart - succ_delay
                # suc_lstarts = [lend - dur + 1 for lend, dur in zip(suc_lends, suc_durations)]
                # earliest_suc_lstart = min(suc_lstarts)

            # if i has precedents
            tasks_precedents = check_if_successors_or_precedents_exist(task_id, successors, precedents)
            latest_pre_finish = 0
            latest_pre_finish_w_delay = num_intervals - 1
            if bool(tasks_precedents):
                prec_durations = [durations[i2] for i2 in tasks_precedents]
                prec_astarts = [actual_starts[i2] for i2 in tasks_precedents]
                succ_delay = succ_delays[task_id]
                latest_pre_finish = max([astart + dur - 1 for astart, dur in zip(prec_durations, prec_astarts)])
                latest_pre_finish_w_delay = latest_pre_finish + succ_delay[0]

            # search for all feasible intervals
            feasible_intervals = []
            for j in range(num_intervals):
                if task_costs[j] < big_value and earliest_suc_lstart_w_delay < j < earliest_suc_lstart - duration + 1 \
                        and latest_pre_finish < j < latest_pre_finish_w_delay:
                    feasible_intervals.append(j)

            try:
                feasible_min_cost = min([task_costs[f] for f in feasible_intervals])
                cheapest_intervals = [f for f in feasible_intervals if task_costs[f] == feasible_min_cost]
                a_start = r.choice(cheapest_intervals) if randomness else cheapest_intervals[0]

                # check max demand constraint
                max_demand_starts = dict()
                temp_profile = household_profile[:]
                try:
                    for d in range(a_start, a_start + duration):
                        temp_profile[d % num_intervals] += power
                except IndexError:
                    print("error")
                temp_max_demand = max(temp_profile)
                while temp_max_demand > max_demand and len(feasible_intervals) > 1:

                    max_demand_starts[a_start] = temp_max_demand
                    feasible_intervals.remove(a_start)

                    feasible_min_cost = min([objective_values[task_id][f] for f in feasible_intervals])
                    feasible_min_cost_indices = [k for k, x in enumerate(objective_values[task_id]) if
                                                 x == feasible_min_cost]
                    # a_start = r.choice(feasible_min_cost_indices)
                    a_start = feasible_min_cost_indices[0]

                    temp_profile = household_profile[:]
                    for d in range(a_start, a_start + duration):
                        temp_profile[d] += power
                    temp_max_demand = max(temp_profile)

                if len(feasible_intervals) == 0 and not max_demand_starts:
                    a_start = min(max_demand_starts, key=max_demand_starts.get)

            except ValueError:
                # print("No feasible intervals left for task", task_id)
                a_start = preferred_starts[task_id]

            actual_starts.append(a_start)

            # obj = 0
            for d in range(a_start, a_start + duration):
                household_profile[d % num_intervals] += power
            # obj += objective_values[task_id][a_start]

        time_scheduling_ogsa = timeit.default_timer() - start_time
        # obj = round(obj, 2)

        return actual_starts, household_profile, time_scheduling_ogsa



# time related parameters
no_intervals = 144
no_periods = 48
no_intervals_periods = int(no_intervals / no_periods)

# household related parameters
# new_households = True
create_new_households = False
no_households = 10
no_full_flex_tasks_min = 5
no_full_flex_tasks_max = 8
no_semi_flex_tasks_min = 0
no_fixed_tasks_min = 0
no_tasks_dependent = 2
maxium_demand_multiplier = no_full_flex_tasks_min + no_semi_flex_tasks_min + no_fixed_tasks_min
care_f_max = 10
care_f_weight = 1

# pricing related parameters
pricing_table_weight = 1
# cost_function_type = "linear"
cost_function_type = "piece-wise"


# solver related parameters
variable_selection = "smallest"
value_choice = "indomain_min"
model_type = "pre"
solver_type = "cp"
solver_name = "gecode"

# external file related parameters
parent_folder = ""
file_cp_pre = parent_folder + 'data/Household-cp-pre.mzn'
file_cp_ini = parent_folder + 'data/Household-cp.mzn'
file_pricing_table = parent_folder + 'data/pricing_table_0.csv'
file_household_area_folder = parent_folder + 'data/'
file_probability = parent_folder + 'data/probability.csv'
file_demand_list = parent_folder + 'data/demands_list.csv'
result_folder = parent_folder + "results/"
file_community_pkl = "community.pkl"
file_community_meta_pkl = "community_meta.pkl"
file_aggregator_pkl = "aggregator.pkl"
file_pricing_table_pkl = "pricing_table.pkl"

# summary related parameters
k0_area = "area"
k0_penalty_weight = "penalty_weight"
k0_households_no = "no_households"
k0_tasks_no = "no_tasks"
k0_cost_type = "cost_function_type"
k0_iteration_no = "no_iterations"

# househole dictionary keys
h_key = "key"
h_psts = "preferred_starts"
h_ests = "earliest_starts"
h_lfs = "latest_finishes"
h_durs = "durations"
h_powers = "powers"
h_cfs = "care_factors"
h_max_cf = "maximum_care_factor"
h_no_precs = "no_precedences"
h_precs = "precedents"
h_succ_delay = "succeeding_delays"
h_max_demand = "maximum_demand"
h_demand_profile = "demand_profile"
h_incon_weight = "inconvenience_cost_weight"


# demand related parameters
k0_household_key = "key"
k0_starts = "start_times"
k0_demand = "demands"
k0_demand_max = "max_demand"
k0_demand_total = "total_demand"
k0_par = "PAR"
k0_final = "final"

# step size
k0_step = "step_size"

# objective related parameters
k0_cost = "cost"
k0_penalty = "inconvenient"
k0_obj = "objective"

# pricing related parameters
k0_prices = "prices"
k0_price_levels = "price_levels"
k0_demand_table = "demand_levels"

# run time related
k0_time = "run_time"
k1_time_scheduling = "rescheduling_time"
k1_time_pricing = "pricing_time"
k1_time_average = "average_run_time"

# k1_interval = "interval"
# k1_period = "period"
k0_algorithm = "algorithm"
k1_minizinc = "minizinc"
k1_ogsa = "ogsa"
k2_before_fw = "scheduling"
k2_after_fw = "pricing"

algorithms = dict()
algorithms[k1_minizinc] = dict()
algorithms[k1_minizinc][k2_before_fw] = k1_minizinc
algorithms[k1_minizinc][k2_after_fw] = f"{k1_minizinc}_fw"
algorithms[k1_ogsa] = dict()
algorithms[k1_ogsa][k2_before_fw] = k1_ogsa
algorithms[k1_ogsa][k2_after_fw] = f"{k1_ogsa}_fw"





from multiprocessing import Pool
import timeit
import random as r
from datetime import timedelta
from fw_ddsm.parameter import *
from minizinc import *


def minizinc_model(model_file, solver, search,
                   objective_values, powers, max_demand, durations,
                   earliest_starts, preferred_starts, latest_ends,
                   successors, precedents, no_precedents, succ_delays,
                   care_factors, prices, inconvenience_cost_weight,
                   num_intervals=no_intervals):
    # problem model
    model = Model(model_file)
    gecode = Solver.lookup(solver)
    model.add_string("solve ")
    if "gecode" in solver:
        model.add_string(":: {} ".format(search))
    model.add_string("minimize obj;")

    ins = Instance(gecode, model)
    num_tasks = len(powers)
    ins["num_intervals"] = num_intervals
    ins["num_tasks"] = num_tasks
    ins["durations"] = durations
    ins["demands"] = powers
    ins["num_precedences"] = no_precedents
    ins["predecessors"] = [p + 1 for p in precedents]
    ins["successors"] = [s + 1 for s in successors]
    ins["prec_delays"] = succ_delays
    ins["max_demand"] = max_demand

    if "ini" in model_type.lower():
        ins["prices"] = prices
        ins["preferred_starts"] = [ps + 1 for ps in preferred_starts]
        ins["earliest_starts"] = [es + 1 for es in earliest_starts]
        ins["latest_ends"] = [le + 1 for le in latest_ends]
        ins["care_factors"] = [cf * inconvenience_cost_weight for cf in care_factors]
    else:
        ins["run_costs"] = objective_values

    # solve problem model
    result = ins.solve(timeout=timedelta(seconds=2))
    # result = ins.solve()

    # process problem solution
    # obj = result.objective
    solution = result.solution.actual_starts
    if "cp" in solver_type:
        actual_starts = [int(a) - 1 for a in solution]
    else:  # "mip" in solver_type:
        actual_starts = [sum([i * int(v) for i, v in enumerate(row)]) for row in solution]
    time = result.statistics["time"].total_seconds()

    optimal_demand_profile = [0] * num_intervals
    for power, duration, a_start, i in zip(powers, durations, actual_starts, range(num_tasks)):
        for t in range(a_start, a_start + duration):
            optimal_demand_profile[t % num_intervals] += power

    return actual_starts, optimal_demand_profile, time


def ogsa(objective_values, big_value, powers, durations, preferred_starts, latest_ends, max_demand,
         successors, precedents, succ_delays, num_intervals=no_intervals, randomness=True):
    start_time = timeit.default_timer()

    def retrieve_successors_or_precedents(list0, prec_or_succ_list1, succ_prec_list2):
        list_r = []
        for l in list0:
            if l in prec_or_succ_list1:
                succ_or_prec_indices = [i2 for i2, k in enumerate(prec_or_succ_list1) if k == l]
                succ_or_prec = [succ_prec_list2[i2] for i2 in succ_or_prec_indices]
                succ_succ_or_prec_prec \
                    = retrieve_successors_or_precedents(succ_or_prec, prec_or_succ_list1, succ_prec_list2)
                list_r.extend(succ_succ_or_prec_prec)
            else:
                list_r.append(l)
        return list_r

    def check_if_successors_or_precedents_exist(checked_task_id, prec_or_succ1, succ_or_prec2):
        succs_succs_or_precs_precs = []
        if checked_task_id in prec_or_succ1:
            indices = [i2 for i2, k in enumerate(prec_or_succ1) if k == checked_task_id]
            succs_or_precs = [succ_or_prec2[i2] for i2 in indices]
            succs_succs_or_precs_precs = retrieve_successors_or_precedents(succs_or_precs, prec_or_succ1, succ_or_prec2)

        return succs_succs_or_precs_precs

    actual_starts = []
    household_profile = [0] * num_intervals
    num_tasks = len(powers)
    for task_id in range(num_tasks):
        power = powers[task_id]
        duration = durations[task_id]
        task_costs = objective_values[task_id]

        # if i has successors
        tasks_successors = check_if_successors_or_precedents_exist(task_id, precedents, successors)
        earliest_suc_lstart_w_delay = 0
        earliest_suc_lstart = num_intervals - 1
        if bool(tasks_successors):
            suc_durations = [durations[i2] for i2 in tasks_successors]
            suc_lends = [latest_ends[i2] for i2 in tasks_successors]
            earliest_suc_lstart = min([lend - dur for lend, dur in zip(suc_lends, suc_durations)])
            # earliest_suc_lstart_w_delay = earliest_suc_lstart - succ_delay
            # suc_lstarts = [lend - dur + 1 for lend, dur in zip(suc_lends, suc_durations)]
            # earliest_suc_lstart = min(suc_lstarts)

        # if i has precedents
        tasks_precedents = check_if_successors_or_precedents_exist(task_id, successors, precedents)
        latest_pre_finish = 0
        latest_pre_finish_w_delay = num_intervals - 1
        if bool(tasks_precedents):
            prec_durations = [durations[i2] for i2 in tasks_precedents]
            prec_astarts = [actual_starts[i2] for i2 in tasks_precedents]
            succ_delay = succ_delays[task_id]
            latest_pre_finish = max([astart + dur - 1 for astart, dur in zip(prec_durations, prec_astarts)])
            latest_pre_finish_w_delay = latest_pre_finish + succ_delay[0]

        # search for all feasible intervals
        feasible_intervals = []
        for j in range(num_intervals):
            if task_costs[j] < big_value and earliest_suc_lstart_w_delay < j < earliest_suc_lstart - duration + 1 \
                    and latest_pre_finish < j < latest_pre_finish_w_delay:
                feasible_intervals.append(j)

        try:
            feasible_min_cost = min([task_costs[f] for f in feasible_intervals])
            cheapest_intervals = [f for f in feasible_intervals if task_costs[f] == feasible_min_cost]
            a_start = r.choice(cheapest_intervals) if randomness else cheapest_intervals[0]

            # check max demand constraint
            max_demand_starts = dict()
            temp_profile = household_profile[:]
            try:
                for d in range(a_start, a_start + duration):
                    temp_profile[d % num_intervals] += power
            except IndexError:
                print("error")
            temp_max_demand = max(temp_profile)
            while temp_max_demand > max_demand and len(feasible_intervals) > 1:

                max_demand_starts[a_start] = temp_max_demand
                feasible_intervals.remove(a_start)

                feasible_min_cost = min([objective_values[task_id][f] for f in feasible_intervals])
                feasible_min_cost_indices = [k for k, x in enumerate(objective_values[task_id]) if
                                             x == feasible_min_cost]
                # a_start = r.choice(feasible_min_cost_indices)
                a_start = feasible_min_cost_indices[0]

                temp_profile = household_profile[:]
                for d in range(a_start, a_start + duration):
                    temp_profile[d] += power
                temp_max_demand = max(temp_profile)

            if len(feasible_intervals) == 0 and not max_demand_starts:
                a_start = min(max_demand_starts, key=max_demand_starts.get)

        except ValueError:
            # print("No feasible intervals left for task", task_id)
            a_start = preferred_starts[task_id]

        actual_starts.append(a_start)

        # obj = 0
        for d in range(a_start, a_start + duration):
            household_profile[d % num_intervals] += power
        # obj += objective_values[task_id][a_start]

    time_scheduling_ogsa = timeit.default_timer() - start_time
    # obj = round(obj, 2)

    return actual_starts, household_profile, time_scheduling_ogsa


# use this for distributed implementation
def schedule_household(household, prices, scheduling_method, model=None, solver=None, search=None):

    def preprocessing():
        max_duration = max(durations)
        # this big cost and big cost * number_tasks need to be smaller than the largest number that the solver can handle
        big_value = max_demand * max_duration * max(prices) + \
                    inconvenience_cost_weight * max_care_factor * num_intervals
        objective_value_matrix = []
        for power, pst, est, lft, dur, cf in zip(powers, preferred_starts, earliest_starts,
                                                 latest_ends, durations, care_factors):
            objective_value_task = []
            for t in range(num_intervals):
                if est <= t <= lft - dur + 1:
                    rc = abs(t - pst) * cf * inconvenience_cost_weight
                    try:
                        rc += sum([prices[j % num_intervals] for j in range(t, t + dur)]) * power
                    except IndexError:
                        print("Error: check the prices.")
                else:
                    rc = big_value
                objective_value_task.append(int(rc))
            objective_value_matrix.append(objective_value_task)
        return objective_value_matrix, big_value

    # read tasks
    key = household[h_key]
    powers = household[h_powers]
    durations = household[h_durs]
    earliest_starts = household[h_ests]
    latest_ends = household[h_lfs]
    preferred_starts = household[h_psts]
    care_factors = household[h_cfs]
    max_care_factor = household[h_max_cf]
    precedents = [x[0] for x in list(household[h_precs].values())]
    successors = list(household[h_precs].keys())
    succ_delays = household[h_succ_delay]  # need to change this format when sending it to the solver
    no_precedents = household[h_no_precs]
    max_demand = household[h_max_demand]
    inconvenience_cost_weight = household[h_incon_weight]

    # read prices
    num_intervals = len(household[h_demand_profile])
    num_periods = len(prices)
    num_intervals_period = int(num_intervals / num_periods)
    if num_periods != num_intervals:
        prices = [int(p) for p in prices for _ in range(num_intervals_period)]
    else:
        prices = [int(p) for p in prices]

    # begin scheduling
    objective_values, big_value = preprocessing()
    if "minizinc" in scheduling_method:
        model = file_cp_pre if model is None else model
        solver = solver_name if solver is None else solver
        search = f"int_search(actual_starts, {variable_selection}, {value_choice}, complete)" \
            if search is None else search
        actual_starts, household_demand_profile, time_scheduling \
            = minizinc_model(model_file=model, solver=solver, search=search,
                             objective_values=objective_values, powers=powers, max_demand=max_demand,
                             durations=durations,earliest_starts=earliest_starts, preferred_starts=preferred_starts,
                             latest_ends=latest_ends, successors=successors, precedents=precedents,
                             no_precedents=no_precedents, succ_delays=succ_delays, care_factors=care_factors,
                             prices=prices, inconvenience_cost_weight=inconvenience_cost_weight,
                             num_intervals=num_intervals)
    else:
        actual_starts, household_demand_profile, time_scheduling \
            = ogsa(objective_values=objective_values, big_value=big_value,
                   powers=powers, durations=durations, preferred_starts=preferred_starts,
                   latest_ends=latest_ends, max_demand=max_demand,
                   successors=successors, precedents=precedents, succ_delays=succ_delays,
                   num_intervals=num_intervals, randomness=False)

    # return results
    penalty_household = sum([abs(pst - ast) * cf for pst, ast, cf
                             in zip(preferred_starts, actual_starts, care_factors)]) * inconvenience_cost_weight

    return {h_key: key, k0_demand: household_demand_profile, k0_starts: actual_starts,
            k0_penalty: penalty_household, k0_time: time_scheduling}


## only needed for iterations on a single machine
def schedule_households(households, prices, num_iteration, scheduling_method, model=None, solver=None, search=None):
    print("Start scheduling households...")
    pool = Pool()
    results = pool.starmap_async(schedule_household,
                                 [(household, prices, scheduling_method, model, solver, search)
                                  for household in households.values()]).get()
    pool.close()
    pool.join()

    aggregate_demand_profile = [0] * no_intervals
    total_inconvenience_cost = 0
    time_scheduling_iteration = 0
    for res in results:
        key = res[h_key]
        households[key][k0_starts][scheduling_method][num_iteration] = res[k0_starts]
        households[key][k0_penalty][scheduling_method][num_iteration] = res[k0_penalty]
        households[key][k0_demand][scheduling_method][num_iteration] = res[k0_demand]

        aggregate_demand_profile = [x + y for x, y in zip(res[k0_demand], aggregate_demand_profile)]
        total_inconvenience_cost += res[k0_penalty]
        time_scheduling_iteration += res[k0_time]

    return households, aggregate_demand_profile, total_inconvenience_cost, time_scheduling_iteration



from pandas import read_csv
import pickle
from math import ceil
from pathlib import Path
from more_itertools import grouper
from fw_ddsm.cfunctions import *
from fw_ddsm.parameter import *


class Aggregator:

    def __init__(self, num_periods=no_periods, cost_function="piece-wise"):
        self.num_periods = num_periods
        self.pricing_table = dict()
        self.cost_function_type = cost_function
        self.aggregator = dict()
        self.start_time_probability = None

    def read(self, read_from_file):
        read_from_file = read_from_file if read_from_file.endswith("/") \
            else read_from_file + "/"
        self.aggregator = self.__existing_aggregator(f"{read_from_file}{file_aggregator_pkl}")
        self.pricing_table = self.__existing_pricing_table(f"{read_from_file}{file_pricing_table_pkl}")
        print("Aggregator is read. ")


    def new(self, normalised_pricing_table_csv, aggregate_preferred_demand_profile, algorithms_options,
            weight=pricing_table_weight, write_to_file_path=None):
        aggregate_preferred_demand_profile = self.__convert_demand_profile(aggregate_preferred_demand_profile)
        maximum_demand_level = max(aggregate_preferred_demand_profile)
        self.pricing_table = self.__new_pricing_table(normalised_pricing_table_csv, maximum_demand_level,
                                                      weight, write_to_file_path)
        print("Pricing table is created. ")
        self.aggregator = self.__new_aggregator(aggregate_preferred_demand_profile, algorithms_options, write_to_file_path)
        print("Aggregator is created. ")


    def update(self, num_iteration, pricing_method, step=None, prices=None, consumption_cost=None,
               demands=None, inconvenience_cost=None, runtime=None, final=None):
        if step is not None:
            self.aggregator[pricing_method][k0_step][num_iteration] = step
        if prices is not None:
            self.aggregator[pricing_method][k0_prices][num_iteration] = prices
        if consumption_cost is not None:
            self.aggregator[pricing_method][k0_cost][num_iteration] = consumption_cost
        if demands is not None:
            demands = self.__convert_demand_profile(demands)
            self.aggregator[pricing_method][k0_demand][num_iteration] = demands
            self.aggregator[pricing_method][k0_demand_max][num_iteration] = max(demands)
            self.aggregator[pricing_method][k0_demand_total][num_iteration] = sum(demands)
            self.aggregator[pricing_method][k0_par][num_iteration] = max(demands) / average(demands)
        if inconvenience_cost is not None:
            self.aggregator[pricing_method][k0_penalty][num_iteration] = inconvenience_cost
        if runtime is not None:
            self.aggregator[pricing_method][k0_time][num_iteration] = runtime
        if final is not None:
            demands = self.__convert_demand_profile(demands)
            self.aggregator[pricing_method][k0_final][k0_demand] = demands
            self.aggregator[pricing_method][k0_final][k0_demand_max] = max(demands)
            self.aggregator[pricing_method][k0_final][k0_par] = max(demands) / average(demands)
            self.aggregator[pricing_method][k0_final][k0_prices] = prices
            self.aggregator[pricing_method][k0_final][k0_cost] = consumption_cost


    def prices_and_cost(self, aggregate_demand_profile):
        prices = []
        consumption_cost = 0
        aggregate_demand_profile = self.__convert_demand_profile(aggregate_demand_profile)

        price_levels = self.pricing_table[k0_price_levels]
        for demand_period, demand_level_period in \
                zip(aggregate_demand_profile, self.pricing_table[k0_demand_table].values()):
            demand_level = list(demand_level_period.values())
            level = bisect_left(demand_level, demand_period)
            if level != len(demand_level):
                price = price_levels[level]
            else:
                price = price_levels[-1]
            prices.append(price)

            if "piece-wise" in self.cost_function_type and level > 0:
                consumption_cost += demand_level[0] * price_levels[0]
                consumption_cost += (demand_period - demand_level[level - 1]) * price
                consumption_cost += sum([(demand_level[i] - demand_level[i - 1]) *
                                         price_levels[i] for i in range(1, level)])
            else:
                consumption_cost += demand_period * price

        consumption_cost = round(consumption_cost, 2)

        return prices, consumption_cost


    def find_step_size(self, num_iteration, demand_profile, inconvenience, pricing_method):
        demand_profile = self.__convert_demand_profile(demand_profile)
        demand_profile_fw_pre = self.aggregator[pricing_method][k0_demand][num_iteration - 1][:]
        inconvenience_fw_pre = self.aggregator[pricing_method][k0_penalty][num_iteration - 1]
        price_fw = self.aggregator[pricing_method][k0_prices][num_iteration - 1][:]
        cost_fw = self.aggregator[pricing_method][k0_cost][num_iteration - 1]

        demand_profile_fw = demand_profile_fw_pre[:]
        inconvenience_fw = inconvenience_fw_pre
        change_of_inconvenience = inconvenience - inconvenience_fw_pre
        demand_profile_changed = [d_n - d_p for d_n, d_p in zip(demand_profile, demand_profile_fw_pre)]
        step_size_final = 0
        min_step_size = 0.001
        gradient = -999
        num_itrs = 0
        while gradient < 0 and step_size_final < 1:
            step_profile = []
            for dp, dn, demand_levels_period in \
                    zip(demand_profile_fw_pre, demand_profile, self.pricing_table[k0_demand_table].values()):
                d_levels = list(demand_levels_period.values())[:-1]
                min_demand_level = min(d_levels)
                max_demand_level = d_levels[-1]
                second_max_demand_level = d_levels[-2]
                if dn < dp < min_demand_level or dp < dn < min_demand_level or dn > dp > second_max_demand_level \
                        or dp > dn > max_demand_level or dn == dp:
                    step = 1
                else:
                    dd = dn - dp
                    dl = find_ge(d_levels, dp) + 0.01 if dd > 0 else find_le(d_levels, dp) - 0.01
                    step = (dl - dp) / dd
                    step = ceil(step * 1000) / 1000
                    step = step if step > min_step_size else 1
                    # step = max(step, min_step_size)
                step_profile.append(step)
            step_size_incr = min(step_profile)

            demand_profile_fw_temp = [d_p + (d_n - d_p) * step_size_incr for d_p, d_n in
                                      zip(demand_profile_fw_pre, demand_profile)]
            price_fw_temp, cost_fw_temp = self.prices_and_cost(demand_profile_fw_temp)
            gradient = sum([d_c * p_fw for d_c, p_fw in
                            zip(demand_profile_changed, price_fw_temp)]) + change_of_inconvenience

            demand_profile_fw_pre = demand_profile_fw_temp[:]
            step_size_final_temp = step_size_final + step_size_incr
            if gradient < 0 and step_size_final_temp < 1:
                step_size_final = step_size_final_temp
                demand_profile_fw = demand_profile_fw_temp[:]
                # print("best step size", best_step_size)
                price_fw = price_fw_temp[:]
                # print("after fw", price_day)
                cost_fw = cost_fw_temp
                inconvenience_fw = inconvenience_fw_pre + step_size_final * change_of_inconvenience
                # print("cost", cost)
                num_itrs += 1

        print(f"Found the best step size {step_size_final} in {num_itrs} iterations at the cost of {cost_fw}")
        return demand_profile_fw, step_size_final, price_fw, cost_fw, inconvenience_fw


    def compute_start_time_probabilities(self, pricing_method):
        prob_dist = []
        history_steps = list(self.aggregator[pricing_method][k0_step].values())
        del history_steps[0]

        for alpha in history_steps:
            if not prob_dist:
                prob_dist.append(1 - alpha)
                prob_dist.append(alpha)
            else:
                prob_dist = [p_d * (1 - alpha) for p_d in prob_dist]
                prob_dist.append(alpha)
        self.start_time_probability = prob_dist.copy()

        return prob_dist


    def __convert_demand_profile(self, aggregate_demand_profile_interval):
        num_intervals = len(aggregate_demand_profile_interval)
        num_intervals_periods = int(num_intervals / self.num_periods)
        aggregate_demand_profile_period = aggregate_demand_profile_interval
        if num_intervals != self.num_periods:
            aggregate_demand_profile_period = [sum(x) for x in
                                                  grouper(num_intervals_periods, aggregate_demand_profile_interval)]

        return aggregate_demand_profile_period


    def  __new_pricing_table(self, normalised_pricing_table_csv, maximum_demand_level, weight=pricing_table_weight,
                             write_to_file_path=None):
    # ---------------------------------------------------------------------- #
    # normalised_pricing_table_csv:
    #       the path of the CSV file of the normalised pricing table
    # demand_level_scalar:
    #       the scalar for rescaling the normalised demand levels
    # ---------------------------------------------------------------------- #

        num_periods = self.num_periods
        csv_table = read_csv(normalised_pricing_table_csv, header=None)
        num_levels = len(csv_table.index)
        demand_level_scalar = maximum_demand_level * weight
        csv_table.loc[num_levels + 1] = [csv_table[0].values[-1] * 10] + [demand_level_scalar * 1.2 for _ in range(num_periods)]

        zero_digit = 100
        pricing_table = dict()
        pricing_table[k0_price_levels] = list(csv_table[0].values)
        pricing_table[k0_demand_table] = dict()
        pricing_table[k0_demand_table] = \
            {period:
                 {level:
                     ceil(csv_table[period + 1].values[level] * demand_level_scalar / zero_digit) * zero_digit
                  for level in range(len(csv_table[period + 1]))}
             for period in range(num_periods)}

        if write_to_file_path is not None:
            write_to_file_path = write_to_file_path if write_to_file_path.endswith("/") \
                else write_to_file_path + "/"
            path = Path(write_to_file_path)
            if not path.exists():
                path.mkdir(mode=0o777, parents=True, exist_ok=False)

            with open(f"{write_to_file_path}{file_pricing_table_pkl}", 'wb+') as f:
                pickle.dump(pricing_table, f, pickle.HIGHEST_PROTOCOL)
            f.close()

        return pricing_table


    def __existing_pricing_table(self, file_path):
        # ---------------------------------------------------------------------- #
        # ---------------------------------------------------------------------- #

        with open(file_path, 'rb') as f:
            pricing_table = pickle.load(f)
        f.close()

        return pricing_table


    def __new_aggregator(self, aggregate_preferred_demand_profile, algorithms_options, write_to_file_path=None):
        max_demand = max(aggregate_preferred_demand_profile)
        total_demand = sum(aggregate_preferred_demand_profile)
        par = round(max_demand / average(aggregate_preferred_demand_profile), 2)

        aggregator = dict()
        aggregator[k0_demand_max] = max_demand
        aggregator[k0_demand] = aggregate_preferred_demand_profile
        for algorithm in algorithms_options.values():
            for alg in algorithm.values():
                if "fw" in alg:
                    aggregator[alg] = dict()
                    aggregator[alg][k0_demand] = dict()
                    aggregator[alg][k0_demand_max] = dict()
                    aggregator[alg][k0_demand_total] = dict()
                    aggregator[alg][k0_par] = dict()
                    aggregator[alg][k0_penalty] = dict()
                    aggregator[alg][k0_final] = dict()
                    aggregator[alg][k0_prices] = dict()
                    aggregator[alg][k0_cost] = dict()
                    aggregator[alg][k0_step] = dict()
                    aggregator[alg][k0_time] = dict()

                    aggregator[alg][k0_demand][0] = aggregate_preferred_demand_profile
                    aggregator[alg][k0_demand_max][0] = max_demand
                    aggregator[alg][k0_demand_total][0] = total_demand
                    aggregator[alg][k0_par][0] = par
                    aggregator[alg][k0_par][0] = par
                    aggregator[alg][k0_penalty][0] =0
                    aggregator[alg][k0_cost][0] = None
                    aggregator[alg][k0_step][0] = 1
                    aggregator[alg][k0_time][0] = 0

        if write_to_file_path is not None:
            write_to_file_path = write_to_file_path if write_to_file_path.endswith("/") \
                else write_to_file_path + "/"
            path = Path(write_to_file_path)
            if not path.exists():
                path.mkdir(mode=0o777, parents=True, exist_ok=False)

            with open(f"{write_to_file_path}{file_aggregator_pkl}", 'wb+') as f:
                pickle.dump(aggregator, f, pickle.HIGHEST_PROTOCOL)
            f.close()

        return aggregator


    def __existing_aggregator(self, file_path):
        # ---------------------------------------------------------------------- #
        # ---------------------------------------------------------------------- #

        with open(file_path, 'rb') as f:
            aggregator = pickle.load(f)
        f.close()

        return aggregator






from bisect import bisect_left, bisect_right


def index(a, x):
    'Locate the leftmost value exactly equal to x'
    i = bisect_left(a, x)
    if i != len(a) and a[i] == x:
        return i
    raise ValueError


def find_lt(a, x):
    'Find rightmost value less than x'
    i = bisect_left(a, x)
    if i:
        return a[i-1]
    raise ValueError


def find_le(a, x):
    'Find rightmost value less than or equal to x'
    i = bisect_right(a, x)
    if i:
        return a[i-1]
    raise ValueError


def find_gt(a, x):
    'Find leftmost value greater than x'
    i = bisect_right(a, x)
    if i != len(a):
        return a[i]
    raise ValueError


def find_ge(a, x):
    'Find leftmost item greater than or equal to x'
    i = bisect_left(a, x)
    if i != len(a):
        return a[i]
    raise ValueError


def find_second_ge(a, x):
    'Find second leftmost item greater than or equal to x'
    i = bisect_left(a, x)
    if i != len(a):
        return a[i + 1]
    raise ValueError


def average(lst):
    return sum(lst) / len(lst)
